#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "lib.h"
#include "objlist.h"
#include "type.h"
#include "symbol.h"
#include "area.h"

/*****************************************************************
 * area_t
 */

area_t area_destroy(area_t area)
{
  if (area != NULL) {
    if (area->symbols != NULL) {
      while (objlist_extract_head(area->symbols) != NULL)
	;
    }

    memset(area, 0, sizeof(*area));
    free(area);
  }

  return NULL;
}

area_t area_create(c_type_t type)
{
  area_t area;

  area = malloc(sizeof(*area));
  if (area == NULL)
    goto err;
  memset(area, 0, sizeof(*area));

  area->type = type;
  area->size = 0;
  area->align = 1;
  area->symbols = objlist_create(NULL);

  return area;

err:
  if (area != NULL)
    area_destroy(area);
  return NULL;
}

void area_print_simple(area_t area, int indent)
{
  objentry_t entry;
  symbol_t symbol;

  indent_print(indent);
  printf("AREA: (%s, size: %d, align: %d)\n",
	 type_get_word(area->type), area->size, area->align);

  indent++;
  for (entry = objlist_get_head(area->symbols);
       !objlist_is_end(area->symbols, entry);
       entry = objentry_get_next(entry)) {
    symbol = objentry_get_obj(entry);
    symbol_print_simple(symbol, indent);
  }
}

void area_print(area_t area, int indent)
{
  objentry_t entry;
  symbol_t symbol;

  indent_print(indent);
  printf("AREA: (%s, size: %d, align: %d)\n",
	 type_get_word(area->type), area->size, area->align);

  indent++;
  for (entry = objlist_get_head(area->symbols);
       !objlist_is_end(area->symbols, entry);
       entry = objentry_get_next(entry)) {
    symbol = objentry_get_obj(entry);
    symbol_print(symbol, indent);
  }
}

/*****************************************************************
 * library
 */

symbol_t area_search_symbol(area_t area, char *name)
{
  objentry_t entry;
  symbol_t symbol;

  for (entry = objlist_get_head(area->symbols);
       !objlist_is_end(area->symbols, entry);
       entry = objentry_get_next(entry)) {
    symbol = objentry_get_obj(entry);
    if (!strcmp(symbol->name, name))
      return symbol;
  }

  return NULL;
}

symbol_t area_insert_symbol(area_t area, symbol_t symbol)
{
  int size, align_size;
  model_t model;

  if (area == NULL)
    return NULL;

  if (symbol->model == NULL)
    return NULL;

  objlist_insert_obj_tail(area->symbols, symbol, symbol_free);
  if (area->type == C_TYPE_UNION) {
    size = model_get_size(symbol->model);
    if (size < 0)
      return NULL;
    symbol->location.type = area->type;
    symbol->location.base.area = area;
    symbol->location.offset = 0;
    symbol->flags |= SYMBOL_FLAG_LOCATED;
    if (area->size < size) {
      area->size = size;
      area->align = size;
    }
  } else if ((symbol->type == C_TYPE_LABEL) &&
	     (symbol->model->type == C_TYPE_FUNCTION) &&
	     (symbol->model->modifiers.number <= 1)) {
    symbol->location.type = area->type;
    symbol->location.base.area = area;
    symbol->location.offset = 0;
    symbol->flags |= (SYMBOL_FLAG_LOCATED | SYMBOL_FLAG_TEXT);
  } else {
    size = model_get_size(symbol->model);
    if (size < 0)
      return NULL;

    model = symbol->model;
    if (symbol->type == C_TYPE_LABEL) {
      model = model_copy(model);
      model_del_modifier(model);
    }
    align_size = model_get_align(model);
    if (align_size < 0)
      return NULL;
    if (align_size > area->align)
      area->align = align_size;

    area->size = ((area->size + align_size - 1) / align_size) * align_size;
    symbol->location.type = area->type;
    symbol->location.base.area = area;
    symbol->location.offset = area->size;
    symbol->flags |= SYMBOL_FLAG_LOCATED;
    area->size += size;
  }

  return symbol;
}
