#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "lib.h"
#include "objlist.h"
#include "type.h"
#include "symbol.h"
#include "value.h"
#include "asm.h"
#include "asm_syntax.h"
#include "asm_symbol.h"

static int reg = 0;

static int reg_reserve(void)
{
  int r;
  r = reg++;
  return r;
}

static int reg_release(void)
{
  int r;
  r = --reg;
  return r;
}

static char *reg_name(int reg)
{
  static char *name[] = {
    "%ebx",
    "0(%esp)", "4(%esp)", "8(%esp)", "12(%esp)",
    "16(%esp)", "20(%esp)", "24(%esp)", "28(%esp)",
  };
  return name[reg];
}

static int reg_stack_size(void)
{
  return sizeof(int) * 8;
}

int asm_symbol_address_get(symbol_t symbol)
{
  asm_section_change_text();

  printf("\t/* ADDRESS GET (%s, %s) */\n", symbol->name, type_get_word(symbol->type));

  saveline(symbol->line);

  switch (symbol->type) {
  case C_TYPE_LABEL:
    printf("\tmov\t$%s, %%eax\n", symbol->name);
    break;
  case C_TYPE_VARIABLE:
    switch (symbol->location.type) {
    case C_TYPE_STATIC:
      printf("\tmov\t$%s, %%eax\n", symbol->name);
      break;
    case C_TYPE_STACK:
      printf("\tlea\t%d(%%esp), %%eax\n",
	     reg_stack_size() + symbol->location.offset);
      break;
    case C_TYPE_STRUCT:
    case C_TYPE_UNION:
      ERREXIT(1);
    case C_TYPE_SYMBOL:
      asm_symbol_address_get(symbol->location.base.symbol);
      printf("\tlea\t%d(%%eax), %%eax\n", symbol->location.offset);
      break;
    case C_TYPE_POINTER:
      asm_symbol_value_get(symbol->location.base.symbol);
      printf("\tlea\t%d(%%eax), %%eax\n", symbol->location.offset);
      break;
    case C_TYPE_NONE:
      printf("\tmov\t$%s, %%eax\n", symbol->name);
      break;
    default:
      ERREXIT(1);
    }
    break;
  default:
    ERREXIT(1);
  }

  return 0;
}

static int asm_symbol_value_op(symbol_t symbol)
{
  symbol_t *arg;
  int i, reg;

  printf("\t/* VALUE OP (%s, %s) */\n",
	 type_get_word(symbol->type), type_get_word(symbol->value->type));

  arg = symbol->value->obj.op.arg;

  switch (symbol->value->type) {
  case C_TYPE_CAST:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[0]);
    break;
  case C_TYPE_ADDRESS:
    asm_symbol_value_set(arg[0]);
    asm_symbol_address_get(arg[0]);
    break;
  case C_TYPE_POINTER:
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t(%%eax), %%eax\n");
    break;
  case C_TYPE_DOT:
    asm_symbol_value_set(arg[0]);
    asm_symbol_address_get(arg[0]);
    printf("\tmov\t%d(%%eax), %%eax\n", arg[1]->location.offset);
    break;
  case C_TYPE_INC: /* ++x */
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%eax\n", arg[0]->name);
    printf("\tinc\t%%eax\n");
    printf("\tmov\t%%eax, %s\n", arg[0]->name);
    break;
  case C_TYPE_DEC: /* --x */
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%eax\n", arg[0]->name);
    printf("\tdec\t%%eax\n");
    printf("\tmov\t%%eax, %s\n", arg[0]->name);
    break;
  case C_TYPE_INCA: /* x++ */
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%%eax, %%edx\n");
    printf("\tmov\t%s, %%eax\n", arg[0]->name);
    printf("\tinc\t%%eax\n");
    printf("\tmov\t%%eax, %s\n", arg[0]->name);
    printf("\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_DECA: /* x-- */
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%%eax, %%edx\n");
    printf("\tmov\t%s, %%eax\n", arg[0]->name);
    printf("\tdec\t%%eax\n");
    printf("\tmov\t%%eax, %s\n", arg[0]->name);
    printf("\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_ADD:
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[1]);
    reg = reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name(reg));
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name(reg));
    reg_release();
    printf("\tadd\t%%edx, %%eax\n");
    break;
  case C_TYPE_SUB:
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[1]);
    reg = reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name(reg));
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name(reg));
    reg_release();
    printf("\tsub\t%%edx, %%eax\n");
    break;
  case C_TYPE_LSHIFT:
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[1]);
    reg = reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name(reg));
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%ecx\n", reg_name(reg));
    reg_release();
    printf("\tshl\t%%cl, %%eax\n");
    break;
  case C_TYPE_RSHIFT:
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[1]);
    reg = reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name(reg));
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%ecx\n", reg_name(reg));
    reg_release();
    printf("\t%s\t%%cl, %%eax\n",
	   (arg[0]->model->flags & MODEL_FLAG_UNSIGNED) ? "shr" : "sar");
    break;
  case C_TYPE_LT:
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[1]);
    reg = reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name(reg));
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name(reg));
    reg_release();
    printf("\tcmp\t%%edx, %%eax\n");
    printf("\tjl\t1f\n");
    printf("\tmov\t$0, %%eax\n");
    printf("\tjmp\t2f\n");
    printf("1:\tmov\t$1, %%eax\n");
    printf("2:\n");
    break;
  case C_TYPE_GT:
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[1]);
    reg = reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name(reg));
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name(reg));
    reg_release();
    printf("\tcmp\t%%edx, %%eax\n");
    printf("\tjg\t1f\n");
    printf("\tmov\t$0, %%eax\n");
    printf("\tjmp\t2f\n");
    printf("1:\tmov\t$1, %%eax\n");
    printf("2:\n");
    break;
  case C_TYPE_LTEQ:
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[1]);
    reg = reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name(reg));
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name(reg));
    reg_release();
    printf("\tcmp\t%%edx, %%eax\n");
    printf("\tjle\t1f\n");
    printf("\tmov\t$0, %%eax\n");
    printf("\tjmp\t2f\n");
    printf("1:\tmov\t$1, %%eax\n");
    printf("2:\n");
    break;
  case C_TYPE_GTEQ:
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[1]);
    reg = reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name(reg));
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name(reg));
    reg_release();
    printf("\tcmp\t%%edx, %%eax\n");
    printf("\tjge\t1f\n");
    printf("\tmov\t$0, %%eax\n");
    printf("\tjmp\t2f\n");
    printf("1:\tmov\t$1, %%eax\n");
    printf("2:\n");
    break;
  case C_TYPE_EQEQ:
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[1]);
    reg = reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name(reg));
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name(reg));
    reg_release();
    printf("\tcmp\t%%edx, %%eax\n");
    printf("\tje\t1f\n");
    printf("\tmov\t$0, %%eax\n");
    printf("\tjmp\t2f\n");
    printf("1:\tmov\t$1, %%eax\n");
    printf("2:\n");
    break;
  case C_TYPE_NEQ:
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[1]);
    reg = reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name(reg));
    asm_symbol_value_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name(reg));
    reg_release();
    printf("\tcmp\t%%edx, %%eax\n");
    printf("\tjne\t1f\n");
    printf("\tmov\t$0, %%eax\n");
    printf("\tjmp\t2f\n");
    printf("1:\tmov\t$1, %%eax\n");
    printf("2:\n");
    break;
  case C_TYPE_EQ:
    asm_symbol_value_set(arg[1]);
    asm_symbol_value_set(arg[0]);
    asm_symbol_value_get(arg[1]);
    reg = reg_reserve();
    printf("\tmov\t%%eax, %s\n", reg_name(reg));
    asm_symbol_address_get(arg[0]);
    printf("\tmov\t%s, %%edx\n", reg_name(reg));
    reg_release();
    printf("\tmov\t%%edx, (%%eax)\n");
    printf("\tmov\t%%edx, %%eax\n");
    break;
  case C_TYPE_FUNCTION:
    if (symbol->value->obj.function.args.number > 0) {
      printf("\tsub\t$0x%x, %%esp\n", symbol->value->obj.function.args.number * 4);
      for (i = 0; i < symbol->value->obj.function.args.number; i++) {
	asm_symbol_value_set(symbol->value->obj.function.args.arg[i]);
	asm_symbol_value_get(symbol->value->obj.function.args.arg[i]);
	printf("\tmov\t%%eax, 0x%x(%%esp)\n", i * 4);
      }
    }
    printf("\tcall\t%s\n", symbol->value->obj.function.symbol->name);
    if (symbol->value->obj.function.args.number > 0) {
      printf("\tadd\t$0x%x, %%esp\n", symbol->value->obj.function.args.number * 4);
    }
    break;
  default:
    ERREXIT(1);
  }

  return 0;
}

int asm_symbol_value_set(symbol_t symbol)
{
  int reg;

  asm_section_change_text();

  printf("\t/* VALUE SET (%s) */\n", type_get_word(symbol->type));

  saveline(symbol->line);

  switch (symbol->type) {
  case C_TYPE_LABEL:
    break;
  case C_TYPE_VARIABLE:
    if ((symbol->value != NULL) && !(symbol->flags & SYMBOL_FLAG_INITFIRST)) {
      asm_symbol_value_op(symbol);
      if (symbol->flags & SYMBOL_FLAG_LOCATED) {
	reg = reg_reserve();
	printf("\tmov\t%%eax, %s\n", reg_name(reg));
	asm_symbol_address_get(symbol);
	printf("\tmov\t%s, %%edx\n", reg_name(reg));
	reg_release();
	printf("\tmov\t%%edx, (%%eax)\n");
      }
    }
    break;
  default:
    ERREXIT(1);
  }

  return 0;
}

int asm_symbol_value_get(symbol_t symbol)
{
  asm_section_change_text();

  printf("\t/* VALUE GET (%s) */\n", type_get_word(symbol->type));

  saveline(symbol->line);

  switch (symbol->type) {
  case C_TYPE_LABEL:
    printf("\tmov\t$%s, %%eax\n", symbol->name);
    break;
  case C_TYPE_VARIABLE:
    if (symbol->flags & SYMBOL_FLAG_CONSTVALUE) {
      switch (symbol->value->type) {
      case C_TYPE_NUMBER:
	printf("\tmov\t$%d, %%eax\n", symbol->value->obj.number);
	break;
      default:
	ERREXIT(1);
      }
      break;
    }
    if (symbol->flags & SYMBOL_FLAG_LOCATED) {
      asm_symbol_address_get(symbol);
      if (!model_is_address(symbol->model)) {
	if (symbol->model->type == C_TYPE_CHAR) {
	  printf("\tmovb\t(%%eax), %%eax\n");
	  break;
	} else if (symbol->model->type == C_TYPE_SHORT) {
	  printf("\tmovs\t(%%eax), %%eax\n");
	  break;
	}
      }
      printf("\tmov\t(%%eax), %%eax\n");
    }
    break;
  default:
    ERREXIT(1);
  }

  return 0;
}

static int asm_symbol_function(symbol_t symbol)
{
  syntax_t syntax;
  objentry_t entry;
  symbol_t s;
  int n;

  asm_section_change_text();

  saveline(symbol->line);

  if (symbol->scope != C_TYPE_STATIC)
    printf("\t.globl\t%s\n", symbol->name);
  printf("\t.type\t%s, @function\n", symbol->name);
  printf("%s:\n", symbol->name);

  if ((symbol->value != NULL) &&
      (symbol->value->type == C_TYPE_BLOCK)) {
    syntax = symbol->value->obj.block.syntax;
    if (syntax->type == C_TYPE_BLOCK) {
      n = 0;
      for (entry = objlist_get_head(symbol->obj.function.args);
	   !objlist_is_end(symbol->obj.function.args, entry);
	   entry = objentry_get_next(entry)) {
	s = objentry_get_obj(entry);
	printf("\tmov\t%d(%%esp), %%eax\n", (n++) * 4);
	printf("\tmov\t%%eax, %d(%%esp)\n",
	       reg_stack_size() + s->location.offset);
      }
      printf("\tsub\t$0x%x, %%esp\n",
	     reg_stack_size() + symbol->obj.function.stack->size);
      asm_syntax_list(syntax->obj.block.syntaxes);
      printf("%s:\n", symbol->obj.function.retlabel->obj.label.name);
      printf("\tadd\t$0x%x, %%esp\n",
	     reg_stack_size() + symbol->obj.function.stack->size);
      printf("\tret\n\n");
    }
  }

  return 0;
}

static int asm_value_array(value_t value)
{
  objentry_t entry;
  symbol_t symbol;

  for (entry = objlist_get_head(value->obj.array.values);
       !objlist_is_end(value->obj.array.values, entry);
       entry = objentry_get_next(entry)) {
    symbol = objentry_get_obj(entry);
    asm_symbol(symbol);
  }

  return 0;
}

static int asm_symbol_label(symbol_t symbol)
{
  asm_section_change_data();

  saveline(symbol->line);

  if (symbol->scope != C_TYPE_STATIC)
    printf("\t.globl\t%s\n", symbol->name);
  printf("\t.align\t%d\n", model_get_align(symbol->model));
  printf("\t.type\t%s, @object\n", symbol->name);
  printf("\t.size\t%s, %d\n", symbol->name, model_get_size(symbol->model));

  printf("%s:\n", symbol->name);

  if (symbol->value == NULL) {
    printf("\t.int\t0");
  } else {
    switch (symbol->value->type) {
    case C_TYPE_STRING:
      printf("\t.string\t\"%s\"\n\n", symbol->value->obj.string.s);
      break;
    case C_TYPE_ARRAY:
      printf("\n");
      asm_value_array(symbol->value);
      break;
    default:
      ERREXIT(1);
    }
  }

  return 0;
}

static int asm_symbol_variable(symbol_t symbol)
{
  asm_section_change_data();

  saveline(symbol->line);

  if (symbol->scope != C_TYPE_STATIC)
    printf("\t.globl\t%s\n", symbol->name);
  printf("\t.align\t%d\n", model_get_align(symbol->model));
  printf("\t.type\t%s, @object\n", symbol->name);
  printf("\t.size\t%s, %d\n", symbol->name, model_get_size(symbol->model));

  printf("%s:\n", symbol->name);

  if (model_is_pointer(symbol->model)) {
    printf("\t.long");
  } else {
    switch (symbol->model->type) {
    case C_TYPE_CHAR:  printf("\t.byte" ); break;
    case C_TYPE_SHORT: printf("\t.short"); break;
    case C_TYPE_INT:   printf("\t.int"  ); break;
    case C_TYPE_LONG:  printf("\t.long" ); break;
    case C_TYPE_STRUCT:
      printf("\n");
      return asm_value_array(symbol->value);
    default:
      ERREXIT(1);
    }
  }

  printf("\t");

  if (symbol->value == NULL) {
    printf("0");
  } else {
    switch (symbol->value->type) {
    case C_TYPE_NUMBER: printf("%d", symbol->value->obj.number); break;
    case C_TYPE_LABEL:  printf("%s", symbol->value->obj.label.symbol->name); break;
    default:
      ERREXIT(1);
    }
  }

  printf("\n\n");

  return 0;
}

int asm_symbol(symbol_t symbol)
{
  saveline(symbol->line);

  switch (symbol->type) {
  case C_TYPE_LABEL:
    if (symbol_is_function(symbol)) {
      asm_symbol_function(symbol);
      break;
    }
    asm_symbol_label(symbol);
    break;
  case C_TYPE_VARIABLE:
    asm_symbol_variable(symbol);
    break;
  default:
    ERREXIT(1);
  }

  return 0;
}
