#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>

#include <nllibc.h>

struct _dirdesc {
  int fd;
  int offset;
  int size;
  char buffer[sizeof(struct dirent) * 4];
  struct dirent dirent;
};

DIR *opendir(const char *filename)
{
  int fd;
  DIR *dirp;

  fd = open(filename, O_RDONLY);
  if (fd < 0)
    return NULL;

  dirp = fdopendir(fd);
  if (dirp == NULL)
    close(fd);

  return dirp;
}

DIR *fdopendir(int fd)
{
  DIR *dirp = NULL;

  dirp = malloc(sizeof(*dirp));
  if (dirp != NULL) {
    memset(dirp, 0, sizeof(*dirp));
    dirp->fd = fd;
    dirp->offset = 0;
    dirp->size = sizeof(dirp->buffer);
  }

  return dirp;
}

struct dirent *readdir(DIR *dirp)
{
#if !defined(__linux__)
  struct dirent *direntp;
#else
  struct linux_dirent *direntp;
#endif
  int r, minsize;

  minsize =
    (((char *)&((struct dirent *)0)->d_reclen) - (char *)0)
    + sizeof(((struct dirent *)0)->d_reclen);

#if !defined(__linux__)
  direntp = (struct dirent *)dirp->buffer;
#else
  direntp = (struct linux_dirent *)dirp->buffer;
#endif

  while (1) {
    if (dirp->offset >= minsize) {
      if (dirp->offset >= direntp->d_reclen)
	break;
    }

    r = getdents(dirp->fd, dirp->buffer + dirp->offset,
		 dirp->size - dirp->offset);
    if (r <= 0)
      return NULL;

    dirp->offset += r;
  }

  memset(&dirp->dirent, 0, sizeof(dirp->dirent));

  dirp->dirent.d_fileno = direntp->d_fileno;
#if defined(__linux__)
  dirp->dirent.d_off    = direntp->d_off;
#endif
  dirp->dirent.d_reclen = direntp->d_reclen;
#if !defined(__linux__)
  dirp->dirent.d_type   = direntp->d_type;
#else
  dirp->dirent.d_type   = *((char *)direntp + direntp->d_reclen - 1);
#endif
#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__) || defined(__DragonFly__) || defined(__APPLE__)
  dirp->dirent.d_namlen = direntp->d_namlen;
#endif
  strncpy(dirp->dirent.d_name, direntp->d_name, sizeof(dirp->dirent.d_name));

  dirp->offset -= direntp->d_reclen;
  memmove(dirp->buffer, dirp->buffer + direntp->d_reclen, dirp->offset);

  return &dirp->dirent;
}

int closedir(DIR *dirp)
{
  return close(fdclosedir(dirp));
}

int fdclosedir(DIR *dirp)
{
  int fd;
  fd = dirp->fd;
  free(dirp);
  return fd;
}

int dirfd(DIR *dirp)
{
  return dirp->fd;
}
